/*
 *  pe_create_view_file.c
 *  NtileCodePredictor
 *
 *  Created by Naoto Morikawa on 2/26/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-03-05 version 0.1 release
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "np_main.h"
#include "np_pdb_data.h"
#include "np_pe_protein.h"
#include "np_print.h"

extern int read_pdb_file(const char * filename);
extern int codechr_to_int(char * code_chr);
extern void print_pe_protein(struct pe_protein * pt_strct);
extern struct np_settings my_settings;		// main.c
extern struct pe_protein my_pe_protein;		// np_load_protein.c


/* HEADER and COMPND of the PDB file */
void print_view_file_header(const char * filename, FILE * fp_udcode)
{
	FILE * fp_pdb;
	char PDBLine[LINE_WIDTH + 1];
	
	if ((fp_pdb = fopen(filename, "r")) == NULL) {
		fprintf(STDERR, "print_view_file_header(): Can't open %s\n",
			filename);
		exit(1);
	}
	fprintf(fp_udcode,
		"----------------------------------------------------------------------------------\n");
	while (fgets(PDBLine, LINE_WIDTH + 1, fp_pdb) != NULL) {
		if (*PDBLine == 'C') {
			if (!strncmp("COMP", PDBLine, 4))
				fprintf(fp_udcode, "|%s|\n", PDBLine);
		} else if (*PDBLine == 'H') {
			if (!strncmp("HEAD", PDBLine, 4))
				fprintf(fp_udcode, "|%s|\n", PDBLine);
		} else if (*PDBLine == '\n') {
			continue;
		} else {
			break;
		}
	}
	fprintf(fp_udcode,
		"----------------------------------------------------------------------------------\n");
	if (fclose(fp_pdb)) {
		fprintf(STDERR, "print_view_file_header(): Can't close %s\n",
			filename);
		abort();
	}
	fprintf(fp_udcode, "  Num. of tiles per residue = %4.1f\n",
		my_settings.tiles_per_res);
	fprintf(fp_udcode, "\n");
	return;
}

void print_view_file_data(struct np_protein * pt_protein, FILE * fp_udcode)
{
	int mod_ct, seq_ct, atm_ct;	// pe_protein (PDB data)
	int chn_ct, aa_ct;		// np_protein (predction)
	
	fprintf(fp_udcode,
		"<PDB_data atoms=\"%d\" residues=\"%d\" sequences=\"%d\" molecules=\"%d\" models=\"%d\">\n",
		my_pe_protein.atm_num, my_pe_protein.res_num, my_pe_protein.seq_num,
		my_pe_protein.mol_num, my_pe_protein.mod_num);
	for (mod_ct = 1; mod_ct <= my_pe_protein.mod_num; mod_ct++) {
		fprintf(fp_udcode, "\n");
		fprintf(fp_udcode, "<model No.=\"%d\">\n", mod_ct);
		chn_ct = 1;
		aa_ct= 1;
		for (seq_ct = my_pe_protein.model[mod_ct].seq_head;
		     seq_ct <= my_pe_protein.model[mod_ct].seq_tail; seq_ct++) {
			fprintf(fp_udcode,
				"<Atom_sequence No.=\"%d\" from=\"%d-th atom\" to=\"%d-th atom\">\n",
				seq_ct, my_pe_protein.sequence[seq_ct].atom_head,
				my_pe_protein.sequence[seq_ct].atom_tail);
			fprintf(fp_udcode,
				"       [  No. ]  ser.  name res chn num  (       x,        y,        z) occu. id\n");
			for (atm_ct = my_pe_protein.sequence[seq_ct].atom_head;
			     atm_ct <= my_pe_protein.sequence[seq_ct].atom_tail;
			     atm_ct++) {
				fprintf(fp_udcode, "<atom> [%6d] %6d ", atm_ct,
					my_pe_protein.atom[atm_ct].serial_num);
				if (my_pe_protein.atom[atm_ct].atom_name == PE_CC)
					fprintf(fp_udcode, " C  ");
				else if (my_pe_protein.atom[atm_ct].atom_name
					 == PE_CA) {
					if (my_settings.compare_enabled == TRUE) {
						if (pt_protein->aa[aa_ct].comp_result == R_FAIL) {
							fprintf(fp_udcode, " CR ");
						} else if (pt_protein->aa[aa_ct].comp_result == R_CONTAINED) {
							fprintf(fp_udcode, " CY ");
						} else if (pt_protein->aa[aa_ct].comp_result == R_NOPREDICTION) {
							fprintf(fp_udcode, " CW ");
						} else {
							fprintf(fp_udcode, " CA ");
						}
					} else {	// accuracy
						if (pt_protein->aa[aa_ct].accuracy == 0.0) {
							fprintf(fp_udcode, " CW ");
						} else if (pt_protein->aa[aa_ct].accuracy < 0.25) {
							fprintf(fp_udcode, " CR ");
						} else if (pt_protein->aa[aa_ct].accuracy < 0.5) {
							fprintf(fp_udcode, " CY ");
						} else {
							fprintf(fp_udcode, " CA ");
						}
					}
					aa_ct++;	// num of amino acids
				} else if (my_pe_protein.atom[atm_ct].atom_name
					 == PE_NN)
					fprintf(fp_udcode, " N  ");
				else if (my_pe_protein.atom[atm_ct].atom_name
					 == PE_RNA)	/* DNA or RNA */
					fprintf(fp_udcode, "    ");
				else
					fprintf(fp_udcode, " UN ");
				if (my_pe_protein.atom[atm_ct].chain_id == ' ') {
					/* To use sscanf() in ProteinViwer */
					my_pe_protein.atom[atm_ct].chain_id = '-';
				}
				fprintf(fp_udcode,
					" %.3s  %c  %4d (%8.3f, %8.3f, %8.3f) %4.3f %.4s </atom>\n",
					my_pe_protein.atom[atm_ct].res_name,
					my_pe_protein.atom[atm_ct].chain_id,
					my_pe_protein.atom[atm_ct].res_seq,
					my_pe_protein.atom[atm_ct].x,
					my_pe_protein.atom[atm_ct].y,
					my_pe_protein.atom[atm_ct].z,
					my_pe_protein.atom[atm_ct].occupancy,
					my_pe_protein.atom[atm_ct].seg_id);
			}	// atom_ct
			fprintf(fp_udcode, "</Atom_sequence>\n");
			if (seq_ct < my_pe_protein.model[mod_ct].seq_tail)
				fprintf(fp_udcode, "\n");
			chn_ct++;
		}	// seq_ct
		fprintf(fp_udcode, "</model>\n");
		break;		// First model only
	}
	fprintf(fp_udcode, "\n");
	fprintf(fp_udcode, "</PDB_data>\n");
	fprintf(fp_udcode, "\n");
	return;
}


#define EXT_LENGTH (5+3+1)	/* ***[.ent|.pdb|.code] => ***_xxxx.udcode + '\0' */

void create_view_file(const char * filename, struct np_protein * pt_protein)
{
	FILE * fp_pdb;
	FILE * fp_udcode;
	char * infile = NULL;		// .ent or .pdb file
	char * outfile = NULL;		// .udcode file (for ProteinViwer)
	char * s;		// ptr to file extension
	int str_size;
	
	// Open infile
	s = strrchr(filename,'.');
	if ( (strcmp(s, ".ent") == 0) || (strcmp(s, ".pdb") == 0)) {
		str_size = (int) strlen(filename);
		infile = (char *) malloc(str_size * sizeof(char));
		strcpy(infile, filename);
	} else if (strcmp(s, ".code") == 0) {
		BOOLEAN file_opened = FALSE;
		
		str_size = (int) strlen(filename);
		infile = (char *) malloc((str_size - 1) * sizeof(char));
		if (infile == NULL) {
			fprintf(STDERR, "create_view_file() : ERROR infile allocation failed\n");
			return;
		}
		strcpy(infile, filename);
		s = strrchr(infile, '.');	// s = ".code"
		strcpy(s+1, "ent\0");
		if ((fp_pdb = fopen(infile, "r")) != NULL) {
			file_opened = TRUE;
			if (fclose(fp_pdb)) {
				fprintf(STDERR, "create_view_file() : Can't close %s\n",
					infile);
				abort();
			}
		}
		if (file_opened == FALSE) {
			strcpy(s+1, "pdb\0");
			if ((fp_pdb = fopen(infile, "r")) != NULL) {
				file_opened = TRUE;
				if (fclose(fp_pdb)) {
					fprintf(STDERR, "create_view_file() : Can't close %s\n",
						infile);
					abort();
				}
			}
		}
		if (file_opened == FALSE) {
			fprintf(STDERR, "create_view_file() : can't open %s\n",
				infile);
			goto LABEL_OUT;
		}
	}
	if (0) fprintf(STDOUT, "PDB FILE : %s\n", infile);

	// Open outfile
	str_size = (int) strlen(filename);
	outfile = (char *) malloc((str_size + EXT_LENGTH) * sizeof(char));
	if (outfile == NULL) {
		fprintf(STDERR, "create_view_file() : ERROR outfile allocation failed\n");
		goto LABEL_OUT;
	}
	strcpy(outfile, filename);
	s = strrchr(outfile, '.');	/* s = ".ent", ".pdb", or ".code" */
	if (my_settings.compare_enabled == TRUE) {
		strcpy(s, "_comp.udcode\0");
	} else {
		strcpy(s, "_accu.udcode\0");
	}
	fprintf(STDOUT, "VIEW FILE : %s\n", outfile);
	if ((fp_udcode = fopen(outfile, "w")) == NULL) {
		fprintf(STDERR, "create_view_file() : can't open %s\n", outfile);
		goto LABEL_OUT;
	}
	
	// Load PDB file
	if (read_pdb_file(infile) == FAIL) {
		fprintf(STDERR, "create_view_file() : ERROR loading PDB file, %s, failed\n", infile);
		goto LABEL_OUT;
	}
	
	// Write atom coordinates
	print_view_file_header(infile, fp_udcode);
	print_view_file_data(pt_protein, fp_udcode);
	
	if (my_pe_protein.model != NULL) free(my_pe_protein.model);
	if (my_pe_protein.atom != NULL) free(my_pe_protein.atom);
	if (my_pe_protein.sequence != NULL) free(my_pe_protein.sequence);
	
	if (fclose(fp_udcode)) {
		fprintf(STDERR, "create_view_file() : Can't close %s\n",
			outfile);
		abort();
	}
LABEL_OUT:
	if (infile != NULL) free(infile);
	if (outfile != NULL) free(outfile);
	return;
}